/*
* COMPONENT_NAME: ldap.clients
*
* ABSTRACT: generic program to modify an entry's RDN using LDAP
*
* ORIGINS: 202,27
*
* (C) COPYRIGHT International Business Machines Corp. 1997, 2004
* All Rights Reserved
* Licensed Materials - Property of IBM
*
* US Government Users Restricted Rights - Use, duplication or
* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
*/ 
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* ldapmodrdn.c - generic program to modify an entry's RDN using LDAP */

#if defined(__OS400__)
#pragma comment (copyright, \
   "5722SS1  (C) Copyright IBM Corp. 1997, 2004. \
   All rights reserved. \
   US Government Users Restricted Rights - \
   Use, duplication or disclosure restricted \
   by GSA ADP Schedule Contract with IBM Corp. \
   Licensed Materials-Property of IBM")
#endif /* OS400 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <ldap.h>

#ifdef LDAP_SSL_MAX
#if defined(__OS400__)
#include <skit40/ibmgsk.h>
#else
#include <ldapssl.h>
#endif
#endif

#if defined( __OS400__ )
#include <QgldHeap.H>
#include <unistd.h>
#include "lber-int.h" /* xlate_local_to_utf8 */
#endif

#ifndef __OS400__ 
/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else 
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( void );
#endif


#ifndef _WIN32
extern int errno;
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *base = NULL;
static char *passwd = "";
static char *passwd_utf8; /* on all platforms other than OS/400 passwd == passwd_utf8 */
static char *ldaphost = "localhost";
static char *mech = NULL;
static int hoplimit = 10;
static int ldapport = LDAP_PORT;
static int doit = LDAP_TRUE;
static int verbose = LDAP_FALSE;
static int contoper = LDAP_FALSE;
static int referrals = LDAP_OPT_ON;
static int DebugLevel = 0;
static int deref = LDAP_DEREF_NEVER;
static LDAP *ld;
static int ldapversion = LDAP_VERSION3;
static FILE *fp = NULL;
static int removeItem = LDAP_FALSE;
static int ssl = 0;
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_dn = NULL;
static char *charset = NULL;
static char *entrydn = NULL;
static char *rdn = NULL;
static char *debug_ptr = NULL;

LDAPControl **serverControls = NULL;

#define safe_realloc( ptr, size ) ( ptr == NULL ? malloc( size ) : \
      realloc( ptr, size ))


/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends modify RDN requests to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s [options] dn newrdn\n", prog);
   fprintf(stderr, "    %s [options] [-i file]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "      dn: distinguished name of entry to rename\n");
   fprintf(stderr, "      newrdn: new relative distinguished name\n");
   fprintf(stderr, "      file: name of input file\n");
   fprintf(stderr, "note:\n");
   fprintf(stderr, "    standard input is used if file is not specified\n");
   fprintf(stderr, "options:\n" );
   /*  please list all options alphabetically */
   fprintf(stderr, "    -c           continuous operation; do not stop processing on error\n");
   fprintf(stderr, "    -C charset   character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level     set debug level in LDAP library\n");
   fprintf(stderr, "    -D dn        bind dn\n");
   fprintf(stderr, "    -h host      LDAP server host name\n");
   fprintf(stderr, "    -i file      read entry modification information from specified file\n");
   fprintf(stderr, "    -f file      read entry modification information from specified file\n");
   fprintf(stderr, "    -k           use server administration control\n");
   fprintf(stderr, "    -K keyfile   file to use for keys\n");
   fprintf(stderr, "    -m mechanism perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M           Manage referral objects as normal entries.\n");
   fprintf(stderr, "    -n           show what would be done but don't actually do it\n");
   fprintf(stderr, "    -N key_name  private key name to use in keyfile\n");
   fprintf(stderr, "    -O maxhops   maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port      LDAP server port number\n");
   fprintf(stderr, "    -P key_pw    keyfile password\n");
   fprintf(stderr, "    -r           remove old entry (move operation)\n");
   fprintf(stderr, "    -R           do not chase referrals\n");
   fprintf(stderr, "    -v           verbose mode\n");
   fprintf(stderr, "    -V version   LDAP protocol version (2 or 3; default is 3)\n");
#ifdef __OS400__

   fprintf(stderr, "    -w password  bind password\n");
#else

   fprintf(stderr, "    -w password  bind password or '?' for non-echoed prompt\n");
#endif

   fprintf(stderr, "    -Z           use a secure ldap connection (SSL)\n");
   exit(1);
}

/*
**      Print a memory error message and exit.
*/
void mem_error(char *varName) {
   printf("%s: Memory allocation error occured.\n", varName);
   exit(1);
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i, port = 0, rc = LDAP_SUCCESS;
   char *optpattern = "cnvrRMZk?h:V:p:D:w:d:i:f:m:O:K:P:N:C:";
   extern char *optarg;
   extern int optind;
   int manageDsa = LDAP_FALSE;
   int adminControl = LDAP_FALSE;

   while (( i = getopt( argc, argv, optpattern )) != EOF ) {
      switch ( i ) {
      case 'V':
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Unsupported version level supplied.\n");
            usage();
         }
         break;
      case 'O':         /* set maximum referral hop count  */
         hoplimit = atoi( optarg );
         break;
      case 'm':         /* use SASL bind mechanism  */
         if (!(mech = strdup ( optarg )))
            mem_error("mech");
         break;
      case 'c':   /* continuous operation mode */
         ++contoper;
         break;
      case 'h':   /* ldap host */
         if (!(ldaphost = strdup( optarg )))
            mem_error("host");
         break;
      case 'D':   /* bind DN */
         if (!(binddn = strdup( optarg )))
            mem_error("bindDN");
         break;
      case 'w':   /* password */
#ifndef __OS400__

         if (optarg && optarg[0] == '?') {
            passwd = getPassword();

         } else
#endif

            if (!(passwd = strdup( optarg )))
               mem_error("password");
         /* OS/400 needs the password in utf8 for CRAM-MD5
         *           and other sasl binds and rebinds */
         passwd_utf8 = passwd;
#ifdef __OS400__

         {
            unsigned long passwd_len = strlen( passwd ) + 1;
            if ( xlate_local_to_utf8( &passwd_utf8, &passwd_len, 0 ) != LDAP_SUCCESS ) {
               passwd_utf8 = NULL;
            }
         }
#endif
         break;
      case 'd':
         DebugLevel = atoi( optarg );
         break;
      case 'i':   /* read from file */
      case 'f':   /* read from file */
         if (( fp = fopen( optarg, "r" )) == NULL ) {
            perror( optarg );
            exit( 1 );
         }
         break;
      case 'p':
         ldapport = atoi( optarg );
         port = 1;
         break;
      case 'n':   /* print adds, don't actually do them */
         doit = 0;
         break;
      case 'R':   /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'M':   /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &serverControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'v':   /* verbose mode */
         verbose++;
         break;
      case 'r':   /* remove old RDN */
         removeItem = LDAP_TRUE;
         break;
      case 'K':
         if (!(keyfile = strdup( optarg )))
            mem_error("keyfile name");
         ssl = 1;
         break;
      case 'P':
         if (!(keyfile_pw = strdup( optarg )))
            mem_error("keyfile password");
         break;
      case 'N':
         if (!(keyfile_dn = strdup( optarg )))
            mem_error("keyfile lable");
         break;
      case 'Z':
         ssl = 1;
         break;
      case 'C':
         if (!(charset = strdup(optarg)))
            mem_error("charset");
         break;
      case 'k':   /* admin control */
         adminControl = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_SERVER_ADMIN_CONTROL_OID, 0 , NULL, LDAP_OPT_ON, &serverControls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case '?':
      default:
         usage();
      }
   }

   if ( (manageDsa | adminControl) && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M and -k options require at least version 3. ");
      if (manageDsa)
         fprintf(stderr, "-M ignored. ");
      if (adminControl)
         fprintf(stderr, "-k ignored.");
      fprintf(stderr, "\n\n");
   }

   /* get dn & rdn */
   if (argc - optind == 2) {
      if (( rdn = strdup( argv[argc - 1] )) == NULL ) {
         perror( "strdup" );
         exit( 1 );
      }
      if (( entrydn = strdup( argv[argc - 2] )) == NULL ) {
         perror( "strdup" );
         exit( 1 );
      }
   } else if ( argc - optind != 0 ) {
      fprintf( stderr, "invalid number of arguments\n");
      usage();
   }

   /* set to standard file input if no additional arguments specified */
   if ((fp == NULL) && (optind >= argc))
      fp = stdin;

   /* Use default SSL port if none specified */
   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT;

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );
}

#ifndef __OS400__ 
/*  Get a password from the user but don't display it. */
char* getPassword( void ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   fputs( "Enter password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}
#endif

/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
#ifdef __OS400__

         if ( mech != NULL )
            *pwp = strdup( passwd_utf8 );
         else
#endif

            *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return LDAP_SUCCESS;
}

/*
** Connect and bind to server.
*/
void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   char *errormsg = NULL;
   char *errormsgctrl = NULL;
   char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **pwdPolicy_controls = NULL;

   /* call ldap_ssl_client_init if V3 and SSL */
   if (ssl && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
#ifdef __OS400__
      if (keyfile) {
         char *xstr = keyfile;
         unsigned long xstrlen = strlen( xstr ) + 1;
         if ( xlate_utf8_to_local( &xstr, &xstrlen, 0 ) != LDAP_SUCCESS ) {
            fprintf( stderr, "keyfile - xlate_utf8_to_local conversion failed\n");
            exit( 1 );
         }
         keyfile = xstr;
      }
      if (keyfile_pw) {
         char *xstr = keyfile_pw;
         unsigned long xstrlen = strlen( xstr ) + 1;
         if ( xlate_utf8_to_local( &xstr, &xstrlen, 0 ) != LDAP_SUCCESS ) {
            fprintf( stderr, "keyfile_pw - xlate_utf8_to_local conversion failed\n");
            exit( 1 );
         }
         keyfile_pw = xstr;
      }
#endif  /* __OS400__ */

      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         exit( 1 ) ;
      }
   }

#ifdef __OS400__
   char *ldaphost_lcp = ldaphost;
   char *keyfile_dn_lcp = keyfile;
   unsigned long len = (unsigned long)(strlen(ldaphost) + 1);
   if (xlate_utf8_to_local(&ldaphost_lcp, &len, 0) != LDAP_SUCCESS) {
      fprintf( stderr, "ldaphost - xlate_utf8_to_local conversion failed\n");
      exit( 1 );
   }

   if (keyfile_dn) {
      len = (unsigned long)(strlen(keyfile_dn) + 1);
      if (xlate_utf8_to_local(&keyfile_dn_lcp, &len, 0) != LDAP_SUCCESS) {
         fprintf( stderr, "keyfile_dn - xlate_utf8_to_local conversion failed\n");
         exit( 1 );
      }
   } else {
      keyfile_dn_lcp = NULL;
   }
#endif  /* __OS400__ */

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n", ldaphost, ldapport,
                   ((keyfile_dn) ? keyfile_dn : "NULL"));
#ifdef __OS400__

         ld = ldap_ssl_init( ldaphost_lcp, ldapport, keyfile_dn_lcp ) ;
#else

         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_dn ) ;
#endif

         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n", ldaphost, ldapport);
#ifdef __OS400__

         if ((ld = ldap_init(ldaphost_lcp, ldapport)) == NULL) {
#else
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
#endif
            perror("ldap_init");
            exit(1);
         }
      }
   }
   else {
      if ( verbose )
         printf( "ldap_open( %s, %d )\n", ldaphost, ldapport );

#ifdef __OS400__

      if (( ld = ldap_open( ldaphost_lcp, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
#else
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
#endif

   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)hoplimit);
   }
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )referrals);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }
#ifdef __OS400__
   ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_OFF);
#endif

   /* call ldap_ssl_start if V2 and SSL */
   if ( ssl && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_dn) ? keyfile_dn : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_dn);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         printf("ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {
      if ( ! mech ) /* Use simple bind */
      {
         /*Set password policy control*/
         if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &pwdPolicy_controls)) != LDAP_SUCCESS) {
            fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, pwdPolicy_controls, NULL, &msgidp);
         /*Free the password policy controls*/
         ldap_controls_free(pwdPolicy_controls);
         rc2 = rc;

         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                                    NULL, &returnedControls, 1 );
         }
         if (rc == LDAP_SUCCESS) {
            if (returnedControls != NULL) {
               ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
            }
         }
         if (rc2 != LDAP_SUCCESS) {
            errormsg = ldap_err2string(rc2);
         } else if (rc != LDAP_SUCCESS) {
            errormsg = ldap_err2string(rc);
         } else if (errcode != LDAP_SUCCESS) {
            errormsg = ldap_err2string(errcode);
         }
         if (controlerr) {
            errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
         }
         if (controlwarn) {
            warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
         }
         if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
            printf("ldap_simple_bind:  ");
            if (errormsg != NULL) {
               printf(errormsg);
               printf(" --- ");
            }
			if (errormsgctrl != NULL) {
               printf(errormsgctrl);
               printf("\n");
			}
			if (warnmsgctrl != NULL) {
               printf(warnmsgctrl, controlres);
               printf("\n");
			}
         } else if (errormsg != NULL) {
            printf("ldap_simple_bind:  %s \n", errormsg);
         }
         if (rc2 != LDAP_SUCCESS) {
            /*
            ** ldap_get_exterror only valuable for ssl communication.
            ** In this example, for LDAP v3, the bind is the first 
            ** instance in which communication actually flows to the
            ** server.  So, if there is an ssl configuration error or
            ** other ssl problem, this will be the first instance where
            ** it will be detected.
            */
            if (ssl) {
               fprintf( stderr, "Attempted communication over SSL.\n");
               fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
            }
            exit( rc2 );
         } else if (errcode != LDAP_SUCCESS) {
            exit( errcode);
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               printf("ldap_sasl_bind_s: %s \n", errormsg);
               printf("ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else {
            if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
               rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
               if (rc != LDAP_SUCCESS ) {
                  errormsg = ldap_err2string(ldap_get_errno(ld));
                  printf("ldap_sasl_bind_s: %s \n", errormsg);
                  printf("ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
                  exit ( rc );
               }
            } else /* other SASL mechanisms */ {
               ber.bv_len = strlen ( passwd );
               ber.bv_val = passwd;
               rc = ldap_sasl_bind_s (ld, binddn, mech, &ber, NULL, NULL, &server_creds);
               if (rc != LDAP_SUCCESS ) {
                  errormsg = ldap_err2string(ldap_get_errno(ld));
                  printf("ldap_sasl_bind_s: %s \n", errormsg);
                  printf("ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
                  exit ( rc );
               }
            }
         }
      }
   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         printf("ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }
}

/*
** Copy or move an entry.
*/
int domodrdn(LDAP *ld, char *dn, char *rdn, int remove) {
   int rc, rc2 = 0, error = 0, errcode = 0;
   LDAPMessage *res;
   char *errmsg = NULL;
   char *errormsg = NULL;
   char *errormsgctrl = NULL;
   char *warnmsgctrl = NULL;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   int msgidp = 0;
   LDAPControl **returnControls = NULL;

   printf( "%s%s %s to %s\n", ((!doit) ? "!" : ""),
           ((remove) ? "moving" : "copying"), dn, rdn);
   if (!doit)
      return LDAP_SUCCESS;

   /*Add the password policy controls*/
   if (ldapversion == LDAP_VERSION3) {
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &serverControls)) != LDAP_SUCCESS) {
         fprintf( stderr, "ldap_modrdn: ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         return rc;
      }
   } else /*version 2 does not support controls*/
      serverControls = NULL;


   rc = ldap_rename( ld, dn, rdn, NULL, remove, serverControls, NULL , &msgidp);

   rc2 = rc; /* Store this result code for later user. */
   if (rc == LDAP_SUCCESS) {
      error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
      rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                              NULL, &returnControls, 1 );
   }
   if (rc == LDAP_SUCCESS) {
      if (returnControls != NULL) {
         ldap_parse_pwdpolicy_response(returnControls, &controlerr, &controlwarn, &controlres);
      }
   }
   if (rc2 != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc2);
      rc = rc2;
   } else if (rc != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc);
   } else if (errcode != LDAP_SUCCESS) {
      errormsg = ldap_err2string(errcode);
      rc = errcode;
   }
   if (controlerr) {
      errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
   }
   if (controlwarn) {
      warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
   }
   if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)){
      printf("ldap_modrdn:  ");
      if (errormsg != NULL) {
         printf(errormsg);
         printf(" --- \n");
      }
	  if (errormsgctrl != NULL) {
         printf(errormsgctrl);
         printf("\n");
	  }
	  if (warnmsgctrl != NULL) {
         printf(warnmsgctrl, controlres);
         printf("\n");
	  }
   } else if ((errormsg != NULL) || (errmsg != NULL)) {
      if (errormsg != NULL)
         if (*errormsg != '\0')
            printf("ldap_modrdn:  %s \n", errormsg);
      if (errmsg != NULL)
         if (*errmsg != '\0')
            printf("ldap_modrdn: additional info: %s\n", errmsg);
   } else if ( verbose ) {
      printf( "rename complete\n" );
   }
   return rc;
}

main( int argc, char **argv ) {
   char *p, buf[ 4096 ];
   int rc, i, authmethod, len, havedn;
   LDAPMod **pmods;
   FILE * cf_fd;

   /* Strip off any path info on program name */
   if ((prog = strrchr(argv[0], '/')) != NULL)
      ++prog;
   else
      prog = argv[0];

#ifdef __OS400__

   prog = "ldapmodrdn";
#endif

   passwd_utf8 = passwd;

   /* parse command line arguments */
   parse_arguments(argc, argv);

   /* connect to server */
   if (doit)
      connect_to_server();

   /* copy or move the item */
   rc = 0;
   if (fp == NULL)
      rc = domodrdn(ld, entrydn, rdn, removeItem);
   else {
      havedn = 0;
      while ((rc == 0 || contoper) && fgets(buf, sizeof(buf), fp) != NULL) {
         if ( *buf != '#' && *buf != '\0' && *buf != '\n' ) { /* comments, blank lines optional, skip */
            /* remove nl */
            len = strlen(buf);
            if ((len > 0) && (buf[len - 1] == '\n'))
               buf[len - 1] = '\0';

            if ( havedn ) { /* have DN, get RDN */
               if (( rdn = strdup( buf )) == NULL ) {
                  perror( "strdup" );
                  exit( 1 );
               }
               rc = domodrdn(ld, entrydn, rdn, removeItem);
               havedn = 0;
            } else if ( !havedn ) { /* don't have DN yet */
               if (( entrydn = strdup( buf )) == NULL ) {
                  perror( "strdup" );
                  exit( 1 );
               }
               ++havedn;
            }
         }
      }
   }

   /* unbind and exit */
   if (doit)
      ldap_unbind(ld);
   if (serverControls)
      ldap_controls_free(serverControls);
   exit(rc);
}

/*
 * NOTE: If you add new prototypes to this file, add an argopt pragma.
 */
#ifdef __OS400_ARG_OPT__
#pragma argopt(domodrdn)
#endif /* __OS400_ARG_OPT__ */
